import pandas as pd
import psycopg2


class DatabaseNotConnectedException(Exception):
    """ Raised when attempting to read from the database without establishing a connection. """


class Database:
    """ Holds a database connection, and transforms query results into Pandas DataFrame """

    def __init__(self, config):
        self.param_dict = {
            "user": config["DB_USER"],
            "password": config["DB_PASSWORD"],
            "host": config["DB_HOST"],
            "port": config["DB_PORT"],
            "database": config["DB_DATABASE"]
        }
        self.debug = config["DEBUG"]
        self.conn = None

    def connect(self):
        """
        Connect to the PostgreSQL database server
        """
        self.conn = None
        try:
            # connect to the PostgreSQL server
            if self.debug:
                print('Connecting to the PostgreSQL database...')
            self.conn = psycopg2.connect(**self.param_dict)
        except (psycopg2.Error, psycopg2.DatabaseError) as error:
            print(error)
        if self.debug:
            print("Connection successful")

    def to_dataframe(self, select_query, column_names, cache=False):
        """ Transform a SELECT query into a pandas dataframe """
        if self.conn is None:
            raise DatabaseNotConnectedException
        cursor = self.conn.cursor()

        if cache is False or select_query not in store:
            try:
                cursor.execute(select_query)
            except (psycopg2.Error, psycopg2.DatabaseError) as error:
                print(f"Error: {error}")
                cursor.close()
                return 1

            # Naturally we get a list of tuples
            tuples = cursor.fetchall()
            cursor.close()

            # We just need to turn it into a pandas dataframe
            df = pd.DataFrame(tuples, columns=column_names)

            if cache is True:
                print(f'df stored {select_query}')
                store[select_query] = df

            return df

        return store[select_query]


store = {}
