import simplejson as sjson
import pandas as pd

from flask import Blueprint
from flask import current_app
from flask_restful import Resource, Api, reqparse

from forecast_source.processing import process_to_json
from optimization_source.pareto_optimal import pareto_optimal
from regioncorrelation_source.regioncorrelation_main import correlation_map, correlation_regionlist, correlation_timeline, correlations_region
from retirement_source.retirement_calculation import retirement_calculation
from wastewater_source.wastewater_main import wastewater_timeline, wastewater_heatmap, wastewater_map
from where2flask.cache import cache
from where2flask.database import Database

bp = Blueprint('rapi', __name__, url_prefix='/rest')
api = Api(bp)


@bp.after_request
def after_request(response):
    response.headers.add('Access-Control-Allow-Origin', '*')
    response.headers.add('Access-Control-Allow-Headers',
                         'Content-Type,Authorization')
    response.headers.add('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE')
    return response


class ParetoOptimal(Resource):
    """ REST call to the pareto_optimal function with all parameters. """

    def __init__(self):
        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument(
            'employees_all',         type=int,   required=True, location='args')
        self.reqparse.add_argument(
            'nv',                    type=int,   required=True, location='args')
        self.reqparse.add_argument(
            'betau',                 type=float, required=True, location='args')
        self.reqparse.add_argument(
            'betav',                 type=float, required=True, location='args')
        self.reqparse.add_argument(
            'contact_rate',          type=float, required=True, location='args')
        self.reqparse.add_argument(
            'home_efficiency_rate',  type=float, required=True, location='args')
        self.reqparse.add_argument(
            'infection_probability', type=float, required=True, location='args')
        self.reqparse.add_argument(
            'test_interval',         type=int,   required=True, location='args')

    def get(self):
        args = self.reqparse.parse_args()
        result = pareto_optimal(
            args['employees_all'],
            args['nv'],
            args['betau'],
            args['betav'],
            args['contact_rate'],
            args['home_efficiency_rate'],
            args['infection_probability'],
            args['test_interval']
        )
        json = sjson.dumps(result)
        return json


@cache.memoize()    # memoize will cache the function call including input parameters
def retirement_cache(*args):
    return retirement_calculation(*args)


class Retirement(Resource):
    """ REST call to the retirement home calculation function with all parameters. """

    def __init__(self):
        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument(
            'm',                 type=int, required=True, location='args')
        self.reqparse.add_argument(
            'n',                 type=int, required=True, location='args')
        self.reqparse.add_argument(
            'kappa',             type=int, required=True, location='args')
        self.reqparse.add_argument(
            'max_g',             type=int, required=True, location='args')
        self.reqparse.add_argument(
            'max_tau',           type=int, required=True, location='args')
        self.reqparse.add_argument(
            'p_time',            type=int, required=True, location='args')
        self.reqparse.add_argument(
            't_time',            type=int, required=True, location='args')
        self.reqparse.add_argument(
            'maximum_workload',  type=int, required=True, location='args')

    def get(self):
        beta = 0.10
        args = self.reqparse.parse_args()

        result = retirement_cache(
            args['m'],
            args['n'],
            args['p_time'] / (8 * 60),
            args['t_time'] / (8 * 60),
            args['maximum_workload'] / 100,
            args['kappa'],
            beta,
            args['max_g'],
            args['max_tau'],
        )

        json = sjson.dumps(result)
        return json


class SaxonyCases(Resource):
    """ Endpoint for getting Saxony incidence data """

    def __init__(self):
        self.db = Database(current_app.config)
        self.db.connect()

    def get(self):
        column_names = ["region", "date", "m", "t", "b", "ch"]
        df = self.db.to_dataframe(
            "SELECT * from view_forecast_saxony_weekly", column_names)
        json_df = process_to_json(df)
        return sjson.dumps(json_df)


class CzechiaCases(Resource):
    """ Endpoint for getting Czechia incidence data """

    def __init__(self):
        # Connection to the Postgresql database
        self.db = Database(current_app.config)
        self.db.connect()

    def get(self):
        column_names = ["region", "date", "m", "t", "b", "ch"]
        df = self.db.to_dataframe(
            "SELECT * from view_forecast_czechia_cases_weekly", column_names)
        json_df = process_to_json(df)
        return sjson.dumps(json_df)


class CzechiaTests(Resource):
    """ Endpoint for getting Czechia test data """

    def __init__(self):
        # Connection to the Postgresql database
        self.db = Database(current_app.config)
        self.db.connect()

    def get(self):
        column_names = ["region", "date", "m", "t", "b", "ch"]
        df = self.db.to_dataframe(
            "SELECT * from view_forecast_czechia_tests_weekly", column_names)
        json_df = process_to_json(df)
        return sjson.dumps(json_df)


class DolnoslaskiCases(Resource):
    """ Endpoint for getting Dolnoslaski incidence data """

    def __init__(self):
        # Connection to the Postgresql database
        self.db = Database(current_app.config)
        self.db.connect()

    def get(self):
        column_names = ["region", "date",
                        "m", "t", "b", "ch"]
        df = self.db.to_dataframe(
            "SELECT * from view_forecast_dolnoslaski_cases_weekly", column_names)
        json_df = process_to_json(df)
        return sjson.dumps(json_df)


class CorrelationRegion(Resource):
    """  Get the correlation information about the selected region"""

    def __init__(self):
        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument(
            'regid',    type=str, required=True, location='args')
        self.reqparse.add_argument(
            'time_diff',    type=int, required=True, location='args')
        self.reqparse.add_argument(
            'time_from',    type=int, required=True, location='args')
        self.reqparse.add_argument(
            'time_to',      type=int, required=True, location='args')

    def get(self):
        args = self.reqparse.parse_args()
        corr_res = correlations_region(args['regid'], args['time_diff'],
                                       args['time_from'], args['time_to'])
        corr_json = sjson.dumps(corr_res)
        return corr_json


class CorrelationMap(Resource):
    """_summary_

    Args:
        Resource (_type_): _description_
    """

    def __init__(self):
        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument(
            'time_from',    type=int, required=True, location='args')
        self.reqparse.add_argument(
            'time_to',      type=int, required=True, location='args')
        self.reqparse.add_argument(
            'mode',      type=str, required=True, location='args')

    def get(self):
        args = self.reqparse.parse_args()
        corr_res = correlation_map(
            args['time_from'], args['time_to'], args['mode'])

        corr_json = sjson.dumps(corr_res)
        return corr_json


class CorrelationTimeline(Resource):
    """Get the list of dates in the data

    """

    def __init__(self):
        self.reqparse = reqparse.RequestParser()

    def get(self):
        args = self.reqparse.parse_args()
        corr_res = correlation_timeline()

        corr_json = sjson.dumps(corr_res)
        return corr_json


class CorrelationRegionList(Resource):
    """Get the list of regions

    """

    def __init__(self):
        self.reqparse = reqparse.RequestParser()

    def get(self):
        args = self.reqparse.parse_args()
        corr_res = correlation_regionlist()

        corr_json = sjson.dumps(corr_res)
        return corr_json

# "view_infection_gemeinde_saxony"


class WastewaterTimeline(Resource):
    """Feeding the Timeline component

    Args:


    Returns:
        json
    """

    def __init__(self):
        # Connection to the Postgresql database
        self.db = Database(current_app.config)
        self.db.connect()

        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument(
            'selection',    type=str, required=False, location='args')
        self.reqparse.add_argument(
            'metric',    type=str, required=True, location='args')

    def get(self):
        args = self.reqparse.parse_args()
        column_names = ["region", "date", "cases"]
        df = self.db.to_dataframe(
            f"SELECT id, date_input, infected from view_infection_gemeinde_saxony WHERE date_input='{args['date']}'", column_names)
        # print(df)
        res = wastewater_timeline(df, args['selection'])

        res_json = sjson.dumps(res)
        return res_json


class WastewaterHeatmap(Resource):
    """For building the heatmap

    Args:
        radius (string): 
        metric (string): 

    Returns:
        json
    """

    def __init__(self):
        # Connection to the Postgresql database
        self.db = Database(current_app.config)
        self.db.connect()

        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument(
            'radius',    type=str, required=True, location='args')
        self.reqparse.add_argument(
            'metric',    type=str, required=True, location='args')

    def get(self):
        args = self.reqparse.parse_args()

        df_cases = False

        if args['radius'] != 'kreis':
            column_names_mun = ["region", "date", "cases"]
            df_cases = self.db.to_dataframe(
                "SELECT region, date_input, infected from view_infection_gemeinde_saxony_weekly WHERE extract(YEAR from date_input) > 2020", column_names_mun, True)
        else:
            column_names_kreis = ["region", "date", "cases"]
            df_cases = self.db.to_dataframe(
                "SELECT region, date, cases from view_casesrki_saxony_weekly WHERE region != '14000' AND extract(YEAR from date) > 2020", column_names_kreis, True)

        column_names_ww = ["plant", "date", "c_e"]
        df_ww = self.db.to_dataframe(
            "SELECT gemeinde_nr, reported_date, c_e from view_wastewater_sachsen_weekly WHERE extract(YEAR from reported_date) > 2020", column_names_ww, True)

        res = wastewater_heatmap(
            df_ww, df_cases, args['radius'], args['metric'])

        res_json = sjson.dumps(res)
        return res_json


# "view_wastewater_sachsen"


class WastewaterMap(Resource):
    """Map component

    Args:
        radius (string): 
        metric (string): 
        date (string):     


    Returns:
        json
    """

    def __init__(self):

        # Connection to the Postgresql database
        self.db = Database(current_app.config)
        self.db.connect()

        self.reqparse = reqparse.RequestParser()
        self.reqparse.add_argument(
            'radius',    type=str, required=True, location='args')
        self.reqparse.add_argument(
            'metric',    type=str, required=True, location='args')
        self.reqparse.add_argument(
            'date',    type=str, required=True, location='args')

        #self.df_cases_kreis['change'] = self.df_cases_kreis['cases'].diff()

        # print(df_plants_data2)

    def get(self):
        args = self.reqparse.parse_args()
        print('loading database')

        df_cases = pd.DataFrame([], columns=['region', 'date', 'cases'])

        if args['radius'] != 'kreis':
            column_names_mun = ["region", "date", "cases"]
            df_cases = self.db.to_dataframe(
                f"SELECT region, date_input, infected from view_infection_gemeinde_saxony_weekly WHERE date_input ='{args['date']}'", column_names_mun, True)
        else:
            column_names_kreis = ["region", "date", "cases"]
            df_cases = self.db.to_dataframe(
                f"SELECT region, date, cases from view_casesrki_saxony_weekly WHERE region != '14000' AND date='{args['date']}'", column_names_kreis, True)

        column_names_ww = ["plant", "date", "c_e"]
        df_ww = self.db.to_dataframe(
            f"SELECT gemeinde_nr, reported_date, c_e from view_wastewater_sachsen_weekly WHERE reported_date='{args['date']}'", column_names_ww, True)

        res = wastewater_map(
            df_ww, df_cases, args['radius'], args['metric'], args['date'])

        res_json = sjson.dumps(res)
        return res_json


api.add_resource(SaxonyCases, "/saxony_cases")
api.add_resource(CzechiaCases, "/czechia_cases")
api.add_resource(CzechiaTests, "/czechia_tests")
api.add_resource(DolnoslaskiCases, "/dolnoslaski_cases")
api.add_resource(ParetoOptimal, "/pareto_optimal")
api.add_resource(Retirement, "/retirement")
api.add_resource(CorrelationRegion, "/correlation_region")
api.add_resource(CorrelationMap, "/correlation_map")
api.add_resource(CorrelationTimeline, "/correlation_timeline")
api.add_resource(CorrelationRegionList, "/correlation_regionlist")

api.add_resource(WastewaterTimeline, "/wastewater_timeline")
api.add_resource(WastewaterHeatmap, "/wastewater_heatmap")
api.add_resource(WastewaterMap, "/wastewater_map")
